#include <math.h>             // Math Functions
#include <stdlib.h>           // Malloc / Free
#include <GL/gl.h>            // OpenGL Library

#include "Gauss.h"

/***  Matrix Gauss Factorization (With Pivoting)  ***/
short int *fattGaussPivot (GLmatrix *Coeff, GLmatrix *Know) {
  GLfloat div;
  GLfloat temp;
  GLfloat max, abs;
  short int *Perm = malloc (Coeff->Row * sizeof (short int));
  short int line;
  short int i, j, k;
  short int dim = Coeff->Row;

  if ((Coeff->Row != Coeff->Col) | (Know->Row != Coeff->Row)) return NULL;

  for (k = 0; k < dim; k++) {
    max = 0.0;
    line = k;
    // Find Pivot
    for (i = k; i < dim; i++) {
      abs = fabsf (Coeff->Data[k + i * dim]);
      if (max < abs) {
        max = abs;
        line = i;
      }
    }
    // Swap Line (In Both Coeff & Know Terms Matrix)
    if (line != k) {
      for (j = k; j < dim; j++) {
        temp = Coeff->Data[k * dim + j];
        Coeff->Data[k * dim + j] = Coeff->Data[line * dim + j];
        Coeff->Data[line * dim + j] = temp;
      }
      for (j = 0; j < Know->Col; j++) {
        temp = Know->Data[k * Know->Col + j];
        Know->Data[k * Know->Col + j] = Know->Data[line * Know->Col + j];
        Know->Data[line * Know->Col + j] = temp;
      }
    }
    // Gauss Factorization
    for (i = (k + 1); i < dim; i++) {
      div = Coeff->Data[i * dim + k] / Coeff->Data[k * dim + k];
      for (j = k; j < dim; j++)
        Coeff->Data[i * dim + j] -= div * Coeff->Data[k * dim + j];
      for (j = 0; j < Know->Col; j++)
        Know->Data[i * Know->Col + j] -= div * Know->Data[k * Know->Col + j];
    }
    Perm[k] = line;
  }

  return Perm;
}

/***  Backward Resolution of Factorized Linear System ***/
void solveBackward (GLmatrix *Coeff, GLmatrix *Know, GLmatrix *Unknow) {
  GLfloat sum;
  short int i, j, k;
  short int dim = Coeff->Row;

  if ((Coeff->Row != Coeff->Col) | (Know->Row != Coeff->Row)) return;

  for (k = 0; k < Unknow->Col; k++) {
    Unknow->Data[(dim - 1) * Unknow->Col + k] =
        Know->Data[(dim - 1) * Know->Col + k] / Coeff->Data[dim * dim - 1];
    for (i = (dim - 2); i >= 0; i--) {
      sum = 0.0;
      for (j = i + 1; j < dim; j++) {
        sum += Coeff->Data[i * dim + j] * Unknow->Data[j * Unknow->Col + k];
      }
      Unknow->Data[i * Unknow->Col + k] =
          (Know->Data[i * Know->Col + k] - sum) / Coeff->Data[i * dim + i];
    }
  }
}

GLmatrix* solveLinearSystem (GLmatrix *Coeff, GLmatrix *Know) {
  short int i, j;
  short int *Perm;
  GLmatrix Temp;
  GLmatrix *Unknow = malloc (sizeof (GLmatrix));

  Unknow->Col = Know->Col; Temp.Col = Know->Col;
  Unknow->Row = Know->Row; Temp.Row = Know->Row;
  Temp.Data = calloc (Temp.Col * Temp.Row, sizeof (GLfloat));
  Unknow->Data = calloc (Unknow->Col * Unknow->Row, sizeof (GLfloat));

  Perm = fattGaussPivot (Coeff, Know);
  solveBackward (Coeff, Know, &Temp);

  // Rearrange Computed Value
  for (i = 0; i < Temp.Row; i++)
    for (j = 0; j < Temp.Col; j++)
      Unknow->Data[i * Unknow->Col + j] = Temp.Data[Perm[i] * Temp.Col + j];

  free (Perm);
  free (Temp.Data);

  return Unknow;
}
