/*
 * Univ. Bologna CG LAB6 camerapath over a landscape
 *
 * based on a Mikael SkiZoWalker's (MoDEL) / France (Skizo@Hol.Fr) demo
 *          and David Bucciarelli (tech.hmw@plus.it)
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#ifdef _WIN32
#include <windows.h>
#endif

#include <GL/glut.h>

#ifndef M_PI
#define M_PI 3.14159265
#endif

#define	FRAME 50

#define FOV 85

#define MAP_SIZE 512
#define MAP_SCALE 1
#define heightMnt 30
static GLfloat terrain[MAP_SIZE * MAP_SIZE];
static GLfloat terraincolor[MAP_SIZE * MAP_SIZE][3];

static int fog = 1;
static int bfcull = 1;
static int poutline = 0;
static int help = 0;

#define WIDTH 640
#define HEIGHT 480
static int scrwidth = WIDTH;
static int scrheight = HEIGHT;

#define OBSSTARTX (MAP_SIZE*MAP_SCALE/2)
#define OBSSTARTY (MAP_SIZE*MAP_SCALE/2)

static float obs[3] = { OBSSTARTX, heightMnt * 0.7, OBSSTARTY };

static float dir[3], v1[2], v2[2];
static float v = 0.0;
static float alpha = 75.0;
static float beta = 90.0;

static float gettime (void)
{
	static clock_t told = 0;
	clock_t tnew, ris;
	tnew = clock ();
	ris = tnew - told;
	told = tnew;
	return (ris / (float) CLOCKS_PER_SEC);
}

static void calcposobs (void)
{
	float alpha1, alpha2;

	dir[0] = sin (alpha * M_PI / 180.0);
	dir[1] = cos (beta * M_PI / 180.0);
	dir[2] = cos (alpha * M_PI / 180.0) * sin (beta * M_PI / 180.0);

	alpha1 = alpha + FOV / 2.0;
	v1[0] = sin (alpha1 * M_PI / 180.0);
	v1[1] = cos (alpha1 * M_PI / 180.0);

	alpha2 = alpha - FOV / 2.0;
	v2[0] = sin (alpha2 * M_PI / 180.0);
	v2[1] = cos (alpha2 * M_PI / 180.0);

	obs[0] += v * dir[0];
	obs[1] += v * dir[1];
	obs[2] += v * dir[2];

	if (obs[1] < 0.0)
		obs[1] = 0.0;

	// pac-man sytle periodicity
	if( obs[0] > MAP_SIZE*MAP_SCALE )
		obs[0] = 0.0;
	if( obs[0] < 0.0 )
		obs[0] = MAP_SIZE*MAP_SCALE;

	if( obs[2] > MAP_SIZE*MAP_SCALE )
		obs[2] = 0.0;
	if( obs[2] < 0.0 )
		obs[2] = MAP_SIZE*MAP_SCALE;
}

static void reshape (int width, int height)
{
	scrwidth = width;
	scrheight = height;
	glViewport (0, 0, (GLint) width, (GLint) height);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (50.0, ((GLfloat) width / (GLfloat) height), MAP_SIZE*MAP_SCALE * 0.001, MAP_SIZE*MAP_SCALE * 0.7);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
}

static void printstring (void *font, char *string)
{
	int len, i;

	len = (int) strlen (string);
	for (i = 0; i < len; i++)
		glutBitmapCharacter (font, string[i]);
}

static void printhelp (void) 
{
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (0.0, 0.0, 0.0, 0.5);
	glRecti (40, 40, 600, 440);
	glDisable (GL_BLEND);

	glColor3f (1.0, 0.0, 0.0);
	glRasterPos2i (300, 420);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "Help");

	glRasterPos2i (60, 390);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "h - Togle Help");
	glRasterPos2i (60, 360);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "f - Togle Fog");
	glRasterPos2i (60, 300);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "p - Wire frame");
	glRasterPos2i (60, 270);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "b - Togle Back face culling");
	glRasterPos2i (60, 240);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "Arrow Keys - Rotate");
	glRasterPos2i (60, 210);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "a - Increase velocity");
	glRasterPos2i (60, 180);
	printstring (GLUT_BITMAP_TIMES_ROMAN_24, "z - Decrease velocity");
}

int roundInt( float a )
{
	return floorf( a>0 ? a+0.5 : a-0.5 );
}

float getHeight(float p[2])
{
	int x = roundInt(p[0]) % MAP_SIZE;
	int y = roundInt(p[1]) % MAP_SIZE;
	return terrain[x + (y * MAP_SIZE)];
}

float* getColor(float p[2])
{
	int x = roundInt(p[0]) % MAP_SIZE;
	int y = roundInt(p[1]) % MAP_SIZE;
	return terraincolor[x + (y * MAP_SIZE)];
}

void set3f(float v[3], float x, float y, float z)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
}

void set2f(float v[2], float x, float y)
{
	v[0] = x;
	v[1] = y;
}


void drawterrain (void)
{
	int X = 0, Y = 0, k = 0;
	float quad[4][2];
	float s;
	int step = 4;

	/* Draw the terrain as opengl quads with stepping some pixels for efficiency. */
	glBegin( GL_QUADS );
	for( X=1; X<MAP_SIZE-step; X+=step )
	{
		for( Y=1; Y<MAP_SIZE-step; Y+=step )
		{
			set2f(quad[0], X      ,Y     );
			set2f(quad[1], X      ,Y+step);
			set2f(quad[2], X+step ,Y+step);
			set2f(quad[3], X+step ,Y     );

			for( k=0 ; k<4 ; k++ )
			{
				glColor3fv(getColor(quad[k]));
				glVertex3f(quad[k][0]*MAP_SCALE, getHeight(quad[k]), quad[k][1]*MAP_SCALE);
			}
		}
	}
	glEnd();

	/* Draw the water as a semi-transparent quad. */
	glDisable (GL_CULL_FACE);
	glEnable (GL_BLEND);
	glBegin (GL_QUADS);
	glColor4f (0.1, 0.7, 1.0, 0.4);
	s = (MAP_SIZE-1)*MAP_SCALE;
	glVertex3f(0*s, heightMnt * 0.3,0*s);
	glVertex3f(1*s, heightMnt * 0.3,0*s);
	glVertex3f(1*s, heightMnt * 0.3,1*s);
	glVertex3f(0*s, heightMnt * 0.3,1*s);
	glEnd ();
	glDisable (GL_BLEND);
	if (bfcull)
		glEnable (GL_CULL_FACE);
	glPopMatrix ();
}

void drawscene (void)
{
	static int count = 0;
	static char frbuf[80];
	float fr;

	glShadeModel (GL_SMOOTH);
	glEnable (GL_DEPTH_TEST);

	if (fog)
		glEnable (GL_FOG);
	else
		glDisable (GL_FOG);

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix ();

	calcposobs ();
	gluLookAt (obs[0], obs[1], obs[2], obs[0] + dir[0], obs[1] + dir[1], obs[2] + dir[2], 0.0, 1.0, 0.0);
	drawterrain();
	glPopMatrix ();

	if ((count % FRAME) == 0)
	{
		fr = gettime ();
		sprintf (frbuf, "Frame rate: %.3f", FRAME / fr);
	}

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_FOG);
	glShadeModel (GL_FLAT);

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (-0.5, 639.5, -0.5, 479.5, -1.0, 1.0);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glColor3f (1.0, 0.0, 0.0);
	glRasterPos2i (10, 10);
	printstring (GLUT_BITMAP_HELVETICA_18, frbuf);

	if (help)
		printhelp ();

	reshape (scrwidth, scrheight);

	glutSwapBuffers ();

	count++;
}

static void key (unsigned char k, int x, int y)
{
	switch (k)
	{
		case 27:
			exit (0);
			break;
		case 'a':
			v += 0.0005;
			break;
		case 'z':
			v -= 0.0005;
			break;
		case 'p':
			if (poutline)
			{
				glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
				poutline = 0;
			}
			else
			{
				glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
				poutline = 1;
			}
			break;
		case 'h':
			help = (!help);
			break;
		case 'f':
			fog = (!fog);
			break;
		case 'b':
			if (bfcull)
			{
				glDisable (GL_CULL_FACE);
				bfcull = 0;
			}
			else
			{
				glEnable (GL_CULL_FACE);
				bfcull = 1;
			}
			break;
	}
}

/* ARGSUSED1 */
static void special (int k, int x, int y)
{
	switch (k)
	{
		case GLUT_KEY_LEFT:
			alpha += 2.0;
			break;
		case GLUT_KEY_RIGHT:
			alpha -= 2.0;
			break;
		case GLUT_KEY_DOWN:
			beta -= 2.0;
			break;
		case GLUT_KEY_UP:
			beta += 2.0;
			break;
	}
}

static void calccolor (GLfloat height, GLfloat c[3])
{
	GLfloat color[4][3] = {
		{1.0, 1.0, 1.0},
		{0.0, 0.8, 0.0},
		{1.0, 1.0, 0.3},
		{0.0, 0.0, 0.8}
	};
	GLfloat fact;

	height = height * (1.0/255.0);

	if (height >= 0.8)
	{
		c[0] = color[0][0];
		c[1] = color[0][1];
		c[2] = color[0][2];
		return;
	}

	if ((height < 0.8) && (height >= 0.5))
	{
		fact = (height - 0.5) * 5.0;
		c[0] = fact * color[0][0] + (1.0 - fact) * color[1][0];
		c[1] = fact * color[0][1] + (1.0 - fact) * color[1][1];
		c[2] = fact * color[0][2] + (1.0 - fact) * color[1][2];
		return;
	}

	if ((height < 0.5) && (height >= 0.4))
	{
		fact = (height - 0.4) * 10.0;
		c[0] = fact * color[1][0] + (1.0 - fact) * color[2][0];
		c[1] = fact * color[1][1] + (1.0 - fact) * color[2][1];
		c[2] = fact * color[1][2] + (1.0 - fact) * color[2][2];
		return;
	}

	if ((height < 0.4) && (height >= 0.3))
	{
		fact = (height - 0.3) * 10.0;
		c[0] = fact * color[2][0] + (1.0 - fact) * color[3][0];
		c[1] = fact * color[2][1] + (1.0 - fact) * color[3][1];
		c[2] = fact * color[2][2] + (1.0 - fact) * color[3][2];
		return;
	}

	c[0] = color[3][0];
	c[1] = color[3][1];
	c[2] = color[3][2];
}

static void loadpic (void)
{
	GLubyte bufferter[MAP_SIZE * MAP_SIZE];
	FILE *FilePic;
	int i, tmp;
	GLenum gluerr;

#ifdef _WIN32
	//if ((FilePic = fopen ("../../texture-map/heightmap.bmp", "r")) == NULL)  //Needs MAP_SIZE 256
	if ((FilePic = fopen ("../../texture-map/Height.bmp", "r")) == NULL)     //Needs MAP_SIZE 512
	//if ((FilePic = fopen ("../../texture-map/Land.bmp", "r")) == NULL)       //Needs MAP_SIZE 512
#else
	//if ((FilePic = fopen ("texture-map/heightmap.bmp", "r")) == NULL)        //Needs MAP_SIZE 256
	if ((FilePic = fopen ("texture-map/Height.bmp", "r")) == NULL)           //Needs MAP_SIZE 512
	//if ((FilePic = fopen ("texture-map/Land.bmp", "r")) == NULL)             //Needs MAP_SIZE 512
#endif
	{
		fprintf (stderr, "Error loading heightfield file\n");
		exit (-1);
	}
	fread (bufferter, MAP_SIZE * MAP_SIZE, 1, FilePic);
	fclose (FilePic);

	for (i = 0; i < (MAP_SIZE * MAP_SIZE); i++)
	{
		//terrain[i] = (bufferter[i] * (heightMnt / (MAP_SIZE-1.0f)));
		terrain[i] = (bufferter[i]/255.0) * heightMnt;
		calccolor ((GLfloat) bufferter[i], terraincolor[i]);
	}
}

static void init (void)
{
	float fogcolor[4] = { 0.6, 0.7, 0.7, 1.0 };

	glClearColor (fogcolor[0], fogcolor[1], fogcolor[2], fogcolor[3]);
	glClearDepth (1.0);
	glDepthFunc (GL_LEQUAL);
	glShadeModel (GL_SMOOTH);
	glEnable (GL_DEPTH_TEST);
	glEnable (GL_CULL_FACE);

	glDisable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable (GL_FOG);
	glFogi (GL_FOG_MODE, GL_EXP2);
	glFogfv (GL_FOG_COLOR, fogcolor);
	glFogf (GL_FOG_DENSITY, 0.01);

	glHint (GL_FOG_HINT, GL_NICEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	reshape (scrwidth, scrheight);
}

int main (int ac, char **av)
{
	glutInitWindowPosition (0, 0);
	glutInitWindowSize (WIDTH, HEIGHT);
	glutInit (&ac, av);

	glutInitDisplayMode (GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

	glutCreateWindow ("Terrain");

	loadpic ();

	init ();

	glutReshapeFunc (reshape);
	glutDisplayFunc (drawscene);
	glutKeyboardFunc (key);
	glutSpecialFunc (special);
	glutIdleFunc (drawscene);

	glutMainLoop ();

	return 0;
}
